---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

-- A D A
with Ada.Unchecked_Deallocation;

-- A S I S
with Asis; use Asis;

-- A S T 2 C F G
with Ast2Cfg.Actuals_For_Traversing;
with Ast2Cfg.Flow_World;
with Ast2Cfg.Flow_Types;
with Ast2Cfg.Pkgs;
with Ast2Cfg.CFGs;
with Ast2Cfg.Nodes;
with Ast2Cfg.Generic_List;
with Ast2Cfg.Generic_Pointer_List;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;

package Ast2Cfg.Transformation is

   procedure Init;
   procedure Final;
   function  Final return Flow_World.World_Object_Ptr;

   procedure Post_Trans_Process(World: in out Flow_World.World_Object);

   procedure Handle_Pre
     (Element :        Asis.Element;
      Control : in out Asis.Traverse_Control;
      State   : in out Actuals_For_Traversing.Traversal_State);

   procedure Handle_Post
     (Element :        Asis.Element;
      Control : in out Asis.Traverse_Control;
      State   : in out Actuals_For_Traversing.Traversal_State);

   procedure Handle_Child_Finish
     (Parent : Asis.Element;
      Child  : Asis.Element;
      Count  : in Integer;
      State  : in out Actuals_For_Traversing.Traversal_State);

   procedure Handle_Unit_Started(Name: String; Class: Asis.Unit_Classes);
   procedure Handle_Unit_Finished(Name: String; Class: Asis.Unit_Classes);

private

   -- ** D A T A  S T R U C T U R E S

   -- * The World we are in
   --   It is aliased because it will be returned to the Library
   World: aliased Flow_World.World_Object;

   -- * The Element Stack Definition
   --   It holds the Path from the Top of the AST to the current Asis.Element
   type Stack_Element is
      record
         Elem:  Asis.Element;
         Child: Integer;
      end record;
   type    Stack_Element_Ptr is access Stack_Element;
   procedure Free_Stack_Elem is new Ada.Unchecked_Deallocation(Stack_Element, Stack_Element_Ptr);
   package Element_Stack     is new Generic_Pointer_List(Item_Pointer => Stack_Element_Ptr,
                                                         Free_Procedure => Free_Stack_Elem);

   -- The Stack Declaration
   EStack: Element_Stack.Object;

   -- * The Node Stack Definition
   --   The top Node in the Stack is always the last Node appended to the CFG
   --   The Stack grows only when the Code branches: the Stack then holds the "if" (..) Node
   --   and (on top) the last Node in the current Branch to be processed
   --   Same goes for Loops: the Stack holds the Loop Head and (on top) the last node in the
   --   Loop to be processed
   package Node_Stack renames Nodes.Node_Class_Ptr_List;

   -- * The Flow Stack Definition
   --   It holds the Path from the Top of the Flow Tree to the current Flow Entityt (CFG/Pkg)
   --   The Nesting of the Flow Structures generates the Flow Tree
   type Flow_Stack_Element is
      record
         Flow:   Flow_Types.Flow_Class_Ptr;
         Elem:   Asis.Element;
         NStack: Node_Stack.Object;
      end record;
   type Flow_Stack_Element_Ptr is access Flow_Stack_Element;
   package Flow_Stack is new Generic_List(Item => Flow_Stack_Element_Ptr);

   -- The Stack Declaration
   FStack: Flow_Stack.Object;

   -- Element freeing Procedure
   procedure Free_Flow_Stack_Elem is new
     Ada.Unchecked_Deallocation(Flow_Stack_Element, Flow_Stack_Element_Ptr);

   -- * The Parameter Stack Definition
   --   This Stack keeps the Path from the Top of a Parameter Tree to the current Element
   --   Such Trees are used (among others) to hold Parameter-Call Hierarchies
   --   (e.g.) in f(g(h(),i()))-like Cases
   package Param_Stack renames Node_Stack;

   -- The Stack Declaration
   PStack: Param_Stack.Object;

   -- * The Transformation Data Container Definition
   --   This Container holds various Data Structures required during the Transformation
   type Trans_Data_Container is
      record
         Decl:        Flow_Types.Declaration_Ptr;                   -- Declaration
         Form_Decls:  Flow_Types.Formal_Decl_Class_Ptr_List.Object; -- Formal Declarations
         Type_Decl:   Flow_Types.Type_Descr_Ptr;                    -- Type Declarations
         Ignore_Mode: Boolean;          -- Ignore several Declaration Types when True
         Names:       UStr_List.Object; -- Names found on the Way
         Rename:      Flow_Types.Alias_Class_Ptr;                   -- Renaming/Instance
         Refs:        Flow_Types.Ref_Class_Ptr_List.Object;         -- Use/With Clauses
         Last_PTree:  Nodes.Node_Class_Ptr;                         -- Last open Param-Tree
      end record;

   -- Countainer Declaration
   Trans_Data: Trans_Data_Container;

   -- ** M E T H O D S

   -- * Element Stack related Methods

   -- Put a new Element onto the Element Stack
   procedure Open_Element (Element: Asis.Element);

   -- Remove the last Element from the Element Stack
   procedure Close_Last_Element;

   -- Remove all Elements from the Stack
   procedure Close_All_Elements;

   -- Return the Container of the top Stack Element
   function Get_Last_Element_Container return Stack_Element_Ptr;

   -- Get the last Element from the Element Stack
   function Get_Last_Element return Asis.Element;

   -- Get the Number of the last Child processed for the current Element
   function Get_Last_Child_Nr return Integer;

   -- Increase the Number of the Childs that have finished for the current Element
   procedure Set_Child_Finished;

   -- * Flow Stack related Methods

   -- Put a new Flow on the Flow Stack
   procedure Open_Flow (Flow_Ptr: in Flow_Types.Flow_Class_Ptr;
                        Elem:     in Asis.Element);

   -- Remove the last Flow from the Flow Stack
   procedure Close_Last_Flow;

   -- Remove all Flows from the Stack
   procedure Close_All_Flows;

   -- Return the Container of the top Stack Element
   function Get_Last_Flow_Container return Flow_Stack_Element_Ptr;

   -- Get Information about the last Flow
   procedure Get_Last_Flow_Info (Flow_Ptr: out Flow_Types.Flow_Class_Ptr;
                                 Pkg_Ptr:  out Pkgs.Pkg_Class_Ptr;
                                 Cfg_Ptr:  out CFGs.CFG_Class_Ptr);

   -- Get the generator Element for the current Flow
   function Get_Last_Flow_Elem return Asis.Element;

   -- * Node Stack related Methods

   -- Put a Node onto the Node Stack
   procedure Add_Node (Node: in Nodes.Node_Class_Ptr);

   -- Remove the last Node from the Node Stack
   function Remove_Last_Node return Nodes.Node_Class_Ptr ;

   -- Get the last Node of the top Flow
   function Get_Last_Node return Nodes.Node_Class_Ptr;

   -- Replace the last Node of the top Flow
   -- > Remove the top Node
   -- > Put the new Node onto the Stack
   -- > Link the second Node from Top to the new top one (and back)
   -- > Returns the old top Node
   function Replace_Last_Node (Node_Ptr: in Nodes.Node_Class_Ptr)
                              return Nodes.Node_Class_Ptr;

   -- Replace the last Node on the top Flow and relink (differently)
   -- > Remove the top node from the Stack
   -- > Put the new Node onto the Stack
   -- > Delete the last successor of the directly specified predecessor
   -- > Link the specified predecessor to the new Node (and back)
   -- > Returns the old top Node
   function Replace_Last_Node_And_Relink (Node_Ptr:      in Nodes.Node_Class_Ptr;
                                          Prev_Node_Ptr: in Nodes.Node_Class_Ptr)
                                         return Nodes.Node_Class_Ptr;

   -- * Parameter Stack related Methods

   -- Put a new Parameter onto the Stack (this one should be the Root of the PTree)
   procedure Open_Param_Root (Param: in Nodes.Node_Class_Ptr);

   -- Put a new Parameter onto the Stack (go one Level deeper into the Parameter Hierarchy)
   procedure Open_Param_Non_Root (Param: in Nodes.Node_Class_Ptr);

   -- Remove the last Parameter from the Stack (go one Level up in the Parameter Hierarchy)
   procedure Close_Last_Param;

   -- Remove all Parameters from the Stack
   procedure Close_All_Params;

   -- Query if there are any active Parameters
   function Param_Tree_Empty return Boolean;

   -- Get the last Parameter from the Stack
   function Get_Last_Param_Node return Nodes.Node_Class_Ptr;

   -- Get the Root of the last Parameter Tree that is *OR WAS* open
   function Get_Last_Param_Root return Nodes.Node_Class_Ptr;

   -- ** A U X I L I A R Y   F U N C T I O N S

   -- Die erste Text-Zeile eines Elements
   function First_Line (Element: Asis.Element) return String;

   -- Add a List of Identifiers to the current Node
   procedure Add_Idents_To_Last_Node (Idents: in out UStr_List.Object);

   -- Add a List of Identifiers to a Type Description Structure
   procedure Add_Idents_To_Type_Description (TDescr: in out Flow_Types.Type_Descr_Ptr;
                                             Idents: in out UStr_List.Object);

   -- Save Identifiers in any open Structure found!
   procedure Save_Identifiers(Idents: in out UStr_List.Object);

end Ast2Cfg.Transformation;
